// SPDX-License-Identifier: Apache-2.0
pragma solidity ^0.8.0;

library CreationCodeHelper {
    /**
     * @notice Generate a creation code that results with a contract with `code` as deployed code.
     *         Generated creation code shall match the one generated by Solidity compiler with an empty constructor.
     * @dev Prepended constructor bytecode consists of:
     *      - 608060405234801561001057600080fd5b50 - store free memory pointer, then check no callvalue is provided
     *      - 61xxxx - push 2 bytes of `code` length
     *      - 806100206000396000f3fe - copy deployed code to memory and return the location of it
     * @param runtimeCode Deployed bytecode to which constructor bytecode will be prepended
     * @return Creation code of a new contract
     */
    function getCreationCodeFor(bytes memory runtimeCode) internal pure returns (bytes memory) {
        return abi.encodePacked(
            hex"608060405234801561001057600080fd5b50",
            hex"61",
            uint16(runtimeCode.length),
            hex"806100206000396000f3fe",
            runtimeCode
        );
    }
}
